/***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU Lesser General Public License as        *
*   published by the Free Software Foundation; either version 2 of the    *
*   License, or (at your option) any later version.                       *
*                                                                         *
***************************************************************************/
/**
@file 
ErodeFault.cpp
@brief
Utility make a map more coherent.
erosion code => Portions Copyright (C) Jason Shankel, 2000
*/

#include "ErodeFault.h"


namespace Ogre
{
    //-----------------------------------------------------------------------
    ErodeFault::ErodeFault()
    {
    }
    //-----------------------------------------------------------------------
    ErodeFault::~ErodeFault()
    { 
    }
    //-----------------------------------------------------------------------
    void ErodeFault::ErodeOneDirection(float * const band, 
                                const int stride, 
                                const uint count, 
                                const float filter) const
    {
	    uint i;
        int j = stride;
	    float v = band[0];
	    for (i = 0; i < count-1; i++)
	    {
		    band[j] = filter*v + (1-filter)*band[j];
		    v = band[j];
		    j += stride;
	    }
    }
    //-----------------------------------------------------------------------
    void ErodeFault::Erode(Real * const field, const uint size,const float filter) const
    {
	    uint i;

	    /*
	    Erode rows left to right
	    */
	    for (i=0;i<size;i++)
	    {
		    ErodeOneDirection( &field[size*i], 1, size, filter);
	    }
    	
	    /*
	    Erode rows right to left
	    */
	    for (i=0;i<size;i++)
	    {
		    ErodeOneDirection (&field[size*i+size-1], -1, size, filter);
	    }

	    /*
	    Erode columns top to bottom
	    */
	    for (i=0;i<size;i++)
	    {
		    ErodeOneDirection (&field[i], (int)size, size, filter);
	    }

	    /*
	    Erode columns bottom to top
	    */
	    for (i=0;i<size;i++)
	    {
		    ErodeOneDirection (&field[size*(size-1)+i], -((int)(size)), size, filter);
	    }
    }
    //-----------------------------------------------------------------------
    void ErodeFault::MakeTerrainFault(Real * const field, const uint size, const uint iterations,
                                         const uint maxDelta, const uint minDelta,
                                         const uint iterationsPerFilter,
                                         const float filter,
                                const float scale) const
    {
        const uint fieldSize = size * size;
	    /*
	    Clear the height field
	    */
        memset (field, 0, fieldSize);

        uint i;
        uint x2, y2;
	    for (i = 0; i < iterations; i++)
	    {
		    /*
		    Calculate the dHeight for this iteration
		    (linear interpolation from maxDelta to minDelta
		    */
		    const Real dHeight = maxDelta - ((maxDelta - minDelta) * i) / iterations;
    		assert (maxDelta - ((maxDelta - minDelta) * i) / iterations > 0);
		    /*
		    Pick two random points on the field for the line
		    (make sure they're not identical)
		    */
		    const uint x1 = rand()%size;
		    const uint y1 = rand()%size;
    		
		    do
		    {
			    x2 = rand()%size;
			    y2 = rand()%size;
		    }while (x2 == x1 && y2 == y1);

    		
		    /*
		    dx1,dy1 is a vector in the direction of the line
		    */
		    const int dx1 = x2 - x1;
		    const int dy1 = y2 - y1;
    		
		    for (x2 = 0; x2 < size; x2++)
            {
		        for (y2 = 0; y2 < size; y2++)
		        {
			        /*
			        dx2,dy2 is a vector from x1,y1 to the candidate point
			        */
			        const int dx2 = x2-x1;
			        const int dy2 = y2-y1;
        			
			        /*
			        if z component of the cross product is 'up', then elevate this point
			        */
			        if (dx2*dy1 - dx1*dy2 > 0)
			        {
				        field[x2+size*y2] += dHeight;
			        }

		        }
            }

		    /*
		    Erode terrain
		    */
		    if (iterationsPerFilter!=0 && (i%iterationsPerFilter)==0)
			    Erode(field, size, filter);
	    }

	    /*
	    Normalize terrain (height field values in the range 0-1)
	    */
	    normalize (field, size, scale);

    }
    //-----------------------------------------------------------------------
    float ErodeFault::RangedRandom(const float v1,const float v2) const
    {
	    return v1 + (v2-v1)*((float)rand())/((float)RAND_MAX);
    }
    //-----------------------------------------------------------------------
    void ErodeFault::normalize (Real * const field, const uint size,
                                const float scale) const
    {
        /*
	    Find the maximum and minimum values in the height field
	    */ 
        const uint fieldSize = size*size;
	    Real maxVal = field[0];
	    Real minVal = field[0];
        uint i;
	    for (i = 1; i < fieldSize; i++)
	    {
		    if (field[i] > maxVal) 
		    {
			    maxVal = field[i];
		    }
		    else if (field[i] < minVal) 
		    {
			    minVal = field[i];
		    }
	    }

	    /*
	    Find the altitude range (dh)
	    */
        const Real minFinal = minVal;
	    if (maxVal <= minFinal) 
            return;
	    const Real inv_dh = scale / (maxVal - minFinal);

	    /*
	    Scale all the values so they fill exactly the range 0-1
	    */
	    for (i=0; i < fieldSize; i++)
	    {
		    field[i] = (field[i] - minFinal) * inv_dh;
            assert (field[i] <= 1.0f && field[i] >= 0.0f);
	    }
    } 
    //-----------------------------------------------------------------------
    void ErodeFault::MakeTerrainPlasma(Real *  const field, 
                                        const uint size, 
                                        const float rough,
                                const float scale) const
    {

	    /*
	    Since the terrain wraps, all 4 "corners" are represented by the value at 0,0,
		    so seeding the heightfield is very straightforward
	    Note that it doesn't matter what we use for a seed value, since we're going to
		    renormalize the terrain after we're done
	    */
	    field[0] = 0;

        const float r = (float)pow (2.0f, -1 * rough);

	    uint i, j, rectSize = size;
	    float dh = (float)rectSize/2;
	    while(rectSize > 0)
	    {
    		
		    /*
		    Diamond step -

		    Find the values at the center of the rectangles by averaging the values at 
		    the corners and adding a random offset:


		    a.....b
		    .     .  
		    .  e  .
		    .     .
		    c.....d   

		    e  = (a+b+c+d)/4 + random

		    In the code below:
		    a = (i,j)
		    b = (ni,j)
		    c = (i,nj)
		    d = (ni,nj)
		    e = (mi,mj)

		    */
    		const uint c_rectSize = rectSize;
    		const float c_dh = dh;

    		const float maxSize = size - (c_rectSize / 2);

		    for (i = 0; i < maxSize; i += c_rectSize)
            {
		        for (j = 0; j < maxSize; j += c_rectSize)
		        {
        			
			        const uint ni = (i+c_rectSize)%size;
			        const uint nj = (j+c_rectSize)%size;

			        const uint mi = (i+c_rectSize/2);
			        const uint mj = (j+c_rectSize/2);
                    

                    assert (i+j*size < size*size);
                    assert (ni+j*size < size*size);
                    assert (ni+nj*size < size*size);
                    assert (i+nj*size < size*size);

                    assert (mi+mj*size < size*size);

			        field[mi+mj*size] = (field[i+j*size] + field[ni+j*size] +
                                         field[i+nj*size] + field[ni+nj*size]) / 4 
                                         + RangedRandom(-c_dh/2, c_dh/2);
		        }
            }
		    /*
		    Square step -

		    Find the values on the left and top sides of each rectangle
		    The right and bottom sides are the left and top sides of the neighboring rectangles,
		    so we don't need to calculate them

		    The height field wraps, so we're never left hanging.  The right side of the last
			    rectangle in a row is the left side of the first rectangle in the row.  The bottom
			    side of the last rectangle in a column is the top side of the first rectangle in
			    the column

                .......
		        .     .
		        .     .
		        .  d  .
		        .     .
		        .     .
		    ......a..g..b
		    .     .     .
		    .     .     .
		    .  e  h  f  .
		    .     .     .
		    .     .     .
		    ......c......

		    g = (d+f+a+b)/4 + random
		    h = (a+c+e+f)/4 + random
    		
		    In the code below:
			    a = (i,j) 
			    b = (ni,j) 
			    c = (i,nj) 
			    d = (mi,pmj) 
			    e = (pmi,mj) 
			    f = (mi,mj) 
			    g = (mi,j)
			    h = (i,mj)

		    */
		    for (i = 0; i < maxSize; i += c_rectSize)
            { 
		        for (j = 0; j < maxSize; j += c_rectSize)
		        {

			        const uint ni = (i+c_rectSize)%size;
			        const uint nj = (j+c_rectSize)%size;

			        const uint mi = (i+c_rectSize/2);
			        const uint mj = (j+c_rectSize/2);

			        const uint pmi = (i-c_rectSize/2+size)%size;
			        const uint pmj = (j-c_rectSize/2+size)%size;

        			
                    assert (i+j*size < size*size);
                    assert (ni+j*size < size*size);
                    assert (i+nj*size < size*size);

                    assert (mi+pmj*size < size*size);
                    assert (mi+mj*size < size*size);
                    assert (pmi+mj*size < size*size);
                    assert (mi+mj*size < size*size);

                    assert (mi+j*size < size*size);
                    assert (i+mj*size < size*size);

			        /*
			        Calculate the square value for the top side of the rectangle
			        */
			        field[mi+j*size] = (field[i+j*size] + field[ni+j*size] + 
                                        field[mi+pmj*size] + field[mi+mj*size]) / 4 
                                        + RangedRandom(-c_dh/2, c_dh/2);

        			
			        /*
			        Calculate the square value for the left side of the rectangle
			        */
			        field[i+mj*size] = (field[i+j*size] + field[i+nj*size] + 
                                        field[pmi+mj*size] + field[mi+mj*size]) / 4 + 
                                        RangedRandom(-c_dh/2, c_dh/2);

		        } 
            }

		    /*
		    Setup values for next iteration
		    At this point, the height field has valid values at each of the coordinates that fall on a rectSize/2 boundary
		    */
		    rectSize /= 2;
		    dh *= r;
	    }
	    /*
	    Normalize terrain so minimum value is 0 and maximum value is 1
	    */
	    normalize(field, size, scale);
    }

    
    //-----------------------------------------------------------------------
    void ErodeFault::MakeTerrainDeposition(Real *  const field, const uint size, const uint jumps,
                                            const uint peakWalk, const uint minParticlesPerJump, 
                                            const uint maxParticlesPerJump,
                                            const float caldera,
                                const float scale) const
    {
	    uint x, y, i, j, p, px, py, minx, miny, maxx, maxy, sx, sy, done;
	    uint dx[] = {0,1,0,size-1,1,1,size-1,size-1};
	    uint dy[] = {1,0,size-1,0,size-1,1,size-1,1};
	    float ch,ph;

	    uint *calderaMap = new uint[size*size];
	    /*
	    Clear out height field
	    */
		memset (field, 0, size*size);
    	
	    for (p = 0; p < jumps; p++)
	    {
    		
		    /*
		    Pick a random spot
		    */
		    x = rand()%size;
		    y = rand()%size;

		    /*
		    px,py will track where we form the caldera
		    */
		    px = x;
		    py = y;
    		
    		
		    /*
		    Determine how many particles we're going to drop
		    */
		    const uint numParticles = (uint) RangedRandom((float)minParticlesPerJump,
                                                          (float)maxParticlesPerJump);
    		
		    for (i = 0; i < numParticles; i++)
		    {

			    /*
			    If it's time to move the drop point, agitate it in a random direction
			    */
			    if (peakWalk != 0 && (i%peakWalk) == 0)
			    {
				    const uint m = rand()%8;
				    x = (x+dx[m]+size)%size;
				    y = (y+dy[m]+size)%size;
			    }


			    /*
			    Drop the particle
			    */
			    field[x+y*size] += 1.0f;
    			
    			
			    /*
			    Now agitate it until it settles
			    */
			    sx = x;
			    sy = y;
			    done = 0;
    			
			    while (done == 0)
			    {
				    done = 1;
    				
				    /*
				    Pick a random neighbor to start inspecting
				    */
				    const uint m = rand();
				    for (j = 0; j < 8 ;j++)
				    {
					    const uint tx = (sx+dx[(j+m)%8])%size;
					    const uint ty = (sy+dy[(j+m)%8])%size;
    					
					    /*
					    If we can move to this neighbor, do it
					    */
					    if (field[tx+ty*size] + 1.0f < field[sx+sy*size])
					    {
						    field[tx+ty*size] += 1.0f;
						    field[sx+sy*size] -= 1.0f;
						    sx = tx;
						    sy = ty;
						    done = 0;
						    break;
					    }
				    }
			    }

			    /*
			    Check to see if our latest point is higher than the caldera point 
			    If so, move the caldera point here
			    */
			    if (field[sx+size*sy] > field[px+size*py])
			    {
				    px = sx;
				    py = sy;
			    }
		    }

		    /*
		    Now that we're done with the peak, invert the caldera
		    ch is the caldera cutoff altitude
		    ph is the height at the caldera start point

		    Note that this technique will not create a caldera at every
			    peak, rather it will create one caldera per jump.  It is
			    possible for the algorithm to generate multiple peaks per jump.
		    */
		    ph = field[px+size*py];
		    ch = ph*(1.0f-caldera);


		    /*
		    We're going to do a flood fill, so we use an array of integers to mark visited locations
		        0 = unmarked, unvisited
		        1 = marked, unvisited
		        2 = marked, visited
		    */
		    memset(calderaMap,0,size*size*sizeof(int));

		    minx = px;
		    maxx = px;
		    miny = py;
		    maxy = py;

		    /*
		    Mark the start location for the caldera
		    */
		    calderaMap[px+size*py] = 1;

		    done = 0;

		    while (done == 0)
		    {
			    done = 1;
			    const uint c_sx = minx;
			    const uint c_sy = miny;
			    const uint tx = maxx;
			    const uint ty = maxy;
    			
			    /*
			    Examine the bounding rectangle looking for unvisited neighbors
			    */
			    for (x = c_sx; x <= tx; x++)
                {
			        for (y = c_sy; y <= ty; y++)
			        {
				        px = (x+size)%size;
				        py = (y+size)%size;

				        /*
				        If this cell is marked but unvisited, check it out
				        */
				        if(calderaMap[px+size*py] == 1)
				        {
					        /*
					        Mark cell as visited
					        */
					        calderaMap[px+size*py] = 2;

					        /*
					        If this cell should be inverted, invert it and inspect neighbors
					        Mark any unmarked, unvisited neighbor
					        Don't invert any cells whose height exceeds our initial caldera height
					        This prevents small peaks from destroying large ones
					        */
					        if(field[px+size*py] > ch && field[px+size*py] <= ph)
					        {
						        done = 0;
						        field[px+size*py] = 2*ch - field[px+size*py];
        					
						        /*
						        Left and right neighbors
						        */
						        px = (px+1)%size;
						        if(calderaMap[px+size*py]==0)
						        {
							        if (x+1 > maxx) maxx = x+1;
							        calderaMap[px+size*py] = 1;
						        }
						        px = (px+size-2)%size;
						        if(calderaMap[px+size*py]==0)
						        {
							        if (x-1 < minx) minx = x-1;
							        calderaMap[px+size*py] = 1;
						        }

						        /*
						        Top and bottom neighbors
						        */
						        px = (x+size)%size;
						        py = (py+1)%size;
						        if(calderaMap[px+size*py] == 0)
						        {
							        if (y+1 > maxy) 
                                        maxy = y+1;
							        calderaMap[px+size*py] = 1;
						        }
						        py = (py+size-2)%size;
						        if(calderaMap[px+size*py] == 0)
						        {
							        if (y-1 < miny) 
                                        miny = y-1;
							        calderaMap[px+size*py] = 1;
						        }
					        }

				        }
			        }
                }
            }

	    }

	    delete [] calderaMap;
    	
	    /*
	    Calderas increase aliasing, so erode terrain with a filter value proportional to the prominence of the caldera
	    */
	    Erode (field, size, caldera);
    	
	    normalize(field, size, scale);
    	
    }
    //-----------------------------------------------------------------------
    void ErodeFault::equalize (Real * const field, const uint size, 
                            const Real hmin, const Real hmax, 
                            const uint maxColor) const
    {
        //maxColor  is 65535 or 255;

        const double divider = 1.0f / (hmax-hmin) * maxColor;
        const uint data_size = size*size;
        uchar *uData = new uchar[data_size];
        uint i;
        for (i = 0; i < data_size; i++ )            
        { 
            uData[i] = uchar  ((field[i]-hmin) * divider);
        }


        uint *freq = new uint[maxColor];
        memset (freq, 0, maxColor*sizeof(uint));
        for (i = 0; i < size; i++ )            
        { 
                freq[uData[i]] += 1;
        }

        uint *Cumulfreq = freq;
        for (i = 1; i < maxColor; i++)
        {           
            Cumulfreq[i] = Cumulfreq[i - 1] + freq[i];
        }

        const double factor = static_cast <double> (hmax - hmin) / data_size;
        uint cur_pos = 0;
        for (i = 0; i < data_size; i++)
        {
            field[i] = Cumulfreq[uData[i]] * factor;           
        }
        delete[] freq;
    } 
}//namespace Ogre
