/*
-----------------------------------------------------------------------------
This source file is part of OGRE
    (Object-oriented Graphics Rendering Engine)
For the latest info, see http://www.ogre3d.org/

Copyright (c) 2000-2005 The OGRE Team
Also see acknowledgements in Readme.html

You may use this sample code for anything you like, it is not covered by the
LGPL like the rest of the engine.
-----------------------------------------------------------------------------
*/

/**
    \file 
        Gui.cpp
    \brief
        An example of CEGUI's features
*/


#define MY_NUM(x) StringConverter::toString(x)
#define VERSION_TEXT "TerraBuild v0.3"
#define TERRAIN_SIZE 512


#include <CEGUI/CEGUIImageset.h>
#include <CEGUI/CEGUISystem.h>
#include <CEGUI/CEGUILogger.h>
#include <CEGUI/CEGUISchemeManager.h>
#include <CEGUI/CEGUIWindowManager.h>
#include <CEGUI/CEGUIWindow.h>
#include <CEGUI/CEGUIcolour.h> 
#include <CEGUI/elements/CEGUICombobox.h>
#include <CEGUI/elements/CEGUIListbox.h>
#include <CEGUI/elements/CEGUIListboxTextItem.h>
#include <CEGUI/elements/CEGUIPushButton.h>
#include <CEGUI/elements/CEGUIScrollbar.h>
#include <CEGUI/elements/CEGUIStaticImage.h>
#include <CEGUI/elements/CEGUIFrameWindow.h>
#include <CEGUI/elements/CEGUICheckbox.h> 
#include <CEGUI/elements/CEGUIEditbox.h> 
#include <CEGUI/elements/CEGUIStaticText.h>
#include <CEGUI/elements/CEGUIRadioButton.h>

#include "OgreCEGUIRenderer.h"
#include "OgreCEGUIResourceProvider.h"

#include "ExampleApplication.h"
#include "CFTerr.h"

CEGUI::MouseButton convertOgreButtonToCegui(int buttonID)
{
    switch (buttonID)
    {
    case MouseEvent::BUTTON0_MASK:
        return CEGUI::LeftButton;
    case MouseEvent::BUTTON1_MASK:
        return CEGUI::RightButton;
    case MouseEvent::BUTTON2_MASK:
        return CEGUI::MiddleButton;
    case MouseEvent::BUTTON3_MASK:
        return CEGUI::X1Button;
    default:
        return CEGUI::LeftButton;
    }
}

class GuiFrameListener : public ExampleFrameListener, public MouseMotionListener, public MouseListener
{
private:
    CEGUI::OgreCEGUIRenderer* mGUIRenderer;
    bool mShutdownRequested;

public:
    // NB using buffered input, this is the only change
    GuiFrameListener(RenderWindow* win, Camera* cam, CEGUI::OgreCEGUIRenderer* renderer)
        : ExampleFrameListener(win, cam, true, true), 
          mGUIRenderer(renderer),
          mShutdownRequested(false)
    {
        mEventProcessor->addMouseMotionListener(this);
        mEventProcessor->addMouseListener(this);
		mEventProcessor->addKeyListener(this);
    }

    /// Tell the frame listener to exit at the end of the next frame
    void requestShutdown(void)
    {
        mShutdownRequested = true;
    }

    bool frameEnded(const FrameEvent& evt)
    {
        if (mShutdownRequested)
            return false;
        else
            return ExampleFrameListener::frameEnded(evt);
    }

    void mouseMoved (MouseEvent *e)
    {
        CEGUI::System::getSingleton().injectMouseMove(
                2.5f * e->getRelX() * mGUIRenderer->getWidth(), 
                2.5f * e->getRelY() * mGUIRenderer->getHeight());
        e->consume();
    }

    void mouseDragged (MouseEvent *e) 
    { 
        mouseMoved(e);
    }

    void mousePressed (MouseEvent *e)
    {
        CEGUI::System::getSingleton().injectMouseButtonDown(
          convertOgreButtonToCegui(e->getButtonID()));
        e->consume();
    }

    void mouseReleased (MouseEvent *e)
    {
        CEGUI::System::getSingleton().injectMouseButtonUp(
          convertOgreButtonToCegui(e->getButtonID()));
        e->consume();
    }

	void mouseClicked(MouseEvent* e) {}
	void mouseEntered(MouseEvent* e) {}
	void mouseExited(MouseEvent* e) {}
    void keyPressed(KeyEvent* e)
    {
        if(e->getKey() == KC_ESCAPE)
        {
            mShutdownRequested = true;
            e->consume();
            return;
        }

        if(e->getKey() == KC_RETURN)
        {
            switchKeyMode();
            e->consume();
            return;
        }

        CEGUI::System::getSingleton().injectKeyDown(e->getKey());
		CEGUI::System::getSingleton().injectChar(e->getKeyChar());
        e->consume();
    }

	void keyReleased(KeyEvent* e)
	{
		CEGUI::System::getSingleton().injectKeyUp(e->getKey());
		e->consume();
	}
	void keyClicked(KeyEvent* e) 
	{
		// Do nothing
		e->consume();
	}

	bool processUnbufferedKeyInput(const FrameEvent& evt)
	{
		// see if switching is on, and you want to toggle 
        if (mInputTypeSwitchingOn && mInputDevice->isKeyDown(KC_RETURN) && mTimeUntilNextToggle <= 0)
		{
			// must be going from immediate keyboard to buffered keyboard
			switchKeyMode();
            mTimeUntilNextToggle = 1;
        }
 
		return ExampleFrameListener::processUnbufferedKeyInput(evt);
	}

	void setKeyInputType(bool itype)
	{
	    mUseBufferedInputKeys = itype;
		mInputDevice->setBufferedInput(mUseBufferedInputKeys, mUseBufferedInputMouse);
	}

};

class GuiApplication : public ExampleApplication
{
private:
    CEGUI::OgreCEGUIRenderer* mGUIRenderer;
    CEGUI::System* mGUISystem;
	CEGUI::Scrollbar* mScaleScr;
	CEGUI::Scrollbar* mOctavesScr;
	CEGUI::Scrollbar* mZoomScr;
	CEGUI::Scrollbar* mOffsetScr;
	CEGUI::Scrollbar* mGainScr;
	CEGUI::StaticText* mScaleTxt;
	CEGUI::StaticText* mOctavesTxt;
	CEGUI::StaticText* mZoomTxt;
	CEGUI::StaticText* mOffsetTxt;
	CEGUI::StaticText* mGainTxt;

	CEGUI::RadioButton* mBasis1;
	CEGUI::RadioButton* mBasis2;
	CEGUI::RadioButton* mBasis3;
	CEGUI::RadioButton* mBasis4;
	CEGUI::RadioButton* mBasis5;
	CEGUI::RadioButton* mBasis6;
	CEGUI::RadioButton* mBasis7;
	CEGUI::RadioButton* mBasis8;
	CEGUI::RadioButton* mBasis9;
	CEGUI::RadioButton* mBasis10;
    CEGUI::RadioButton* mBasis11;

	CEGUI::PushButton* mUpBtn;
	CEGUI::PushButton* mDownBtn;
	CEGUI::PushButton* mLeftBtn;
	CEGUI::PushButton* mRightBtn;
	CEGUI::Editbox* mSeedTxt;
	CEGUI::FrameWindow* FileWnd;
	CEGUI::FrameWindow* HelpWnd;
	CEGUI::Editbox* mFilenameTxt;
	CEGUI::StaticText* mSaveinfoTxt;

	CFTerr *testft;

public:
    GuiApplication()
      : mGUIRenderer(0),
        mGUISystem(0),
        testft (0)
    {
	}

    ~GuiApplication()
    {
        delete mGUISystem;
        delete mGUIRenderer;
		delete testft;
    }

protected:
	//------------------------------------------------------------------------------------------------
	void setupResources(void)
	{
		ExampleApplication::setupResources();

		ResourceGroupManager *rsm = ResourceGroupManager::getSingletonPtr();
		StringVector groups = rsm->getResourceGroups();        
		if (std::find(groups.begin(), groups.end(), String("OgreOde")) == groups.end())
		{
			rsm->createResourceGroup("OgreOde");
			try
			{
				rsm->addResourceLocation("../../../../../ogreaddons/Terrabuild/Media","FileSystem", "Terrabuild");
			}
			catch (Ogre::Exception& e)
			{
				String error = e.getFullDescription();
				rsm->addResourceLocation("../../../Terrabuild/Media","FileSystem", "Terrabuild");
			}
		}

	}

    // Just override the mandatory create scene method
    void createScene(void)
    {
        // Set ambient light
        mSceneMgr->setAmbientLight(ColourValue(0.5, 0.5, 0.5));

        // Create a sky dome
        mSceneMgr->setSkyDome(true, "Examples/CloudySky", 5, 8);

        // Give it a little ambiance with lights
		Light* l;
        l = mSceneMgr->createLight("SunLight");
		l->setType( Light::LT_DIRECTIONAL);
		l->setDirection(-1.0f, -1.0f, 0.0f);
        l->setDiffuseColour(1.0, 1.0, 1.0);

        // setup GUI system
        mGUIRenderer = new CEGUI::OgreCEGUIRenderer(mWindow, 
            Ogre::RENDER_QUEUE_OVERLAY, false, 3000, mSceneMgr);
        mGUISystem = new CEGUI::System(mGUIRenderer);
        CEGUI::Logger::getSingleton().setLoggingLevel(CEGUI::Informative);

		// Terrain mesh
		testft = new CFTerr;
		//testft->NewTerrain("plane01", 50);		// Preview terrain, not in use yet
		testft->NewTerrain("plane02", mSceneMgr, TERRAIN_SIZE);		// Full detail terrain
		//Entity *tent = mSceneMgr->getEntity("plane01");
		//tent->setVisible(false);

        // Position the camera
        mCamera->setPosition(700,500,700);
        mCamera->lookAt(-50,0,0);

        // load scheme and set up defaults
        CEGUI::SchemeManager::getSingleton().loadScheme(
                (CEGUI::utf8*)"TaharezLookSkin.scheme");
        mGUISystem->setDefaultMouseCursor(
                (CEGUI::utf8*)"TaharezLook", (CEGUI::utf8*)"MouseArrow");
        mGUISystem->setDefaultFont((CEGUI::utf8*)"BlueHighway-10");

        CEGUI::Window* sheet = 
            CEGUI::WindowManager::getSingleton().loadWindowLayout(
                (CEGUI::utf8*)"TerraBuild_01.xml"); 
        mGUISystem->setGUISheet(sheet);

		
		CEGUI::WindowManager& wmgr = CEGUI::WindowManager::getSingleton();
        CEGUI::FrameWindow* PropWnd = (CEGUI::FrameWindow*) wmgr.getWindow((CEGUI::utf8*)"PropWnd");
		//PropWnd->setCaptionColour(CEGUI::colour::colour(0xffff0000));
        FileWnd = (CEGUI::FrameWindow*) wmgr.getWindow((CEGUI::utf8*)"FileWnd");
		FileWnd->toggleRollup();
		// Version text
		FileWnd->setText(VERSION_TEXT);

		HelpWnd = (CEGUI::FrameWindow*) wmgr.getWindow((CEGUI::utf8*)"HelpWnd");
		HelpWnd->toggleRollup();
		HelpWnd->moveToBack();
		HelpWnd->hide();			// This window not in use yet

	
		setupGuiObjects();
		setupEventHandlers();
    }

    // Create new frame listener
    void createFrameListener(void)
    {
        mFrameListener= new GuiFrameListener(mWindow, mCamera, mGUIRenderer);
        mRoot->addFrameListener(mFrameListener);
    }

	void setupGuiObjects(void)
	{
		CEGUI::WindowManager& wmgr = CEGUI::WindowManager::getSingleton();
		mScaleScr = static_cast<CEGUI::Scrollbar*>(wmgr.getWindow((CEGUI::utf8*)"ScaleScr"));
		mOctavesScr = static_cast<CEGUI::Scrollbar*>(wmgr.getWindow((CEGUI::utf8*)"OctavesScr"));
		mZoomScr = static_cast<CEGUI::Scrollbar*>(wmgr.getWindow((CEGUI::utf8*)"ZoomScr"));
		mOffsetScr = static_cast<CEGUI::Scrollbar*>(wmgr.getWindow((CEGUI::utf8*)"OffsetScr"));
		mGainScr = static_cast<CEGUI::Scrollbar*>(wmgr.getWindow((CEGUI::utf8*)"GainScr"));
		mScaleTxt = static_cast<CEGUI::StaticText*>(wmgr.getWindow((CEGUI::utf8*)"ScaleTxt"));
		mOctavesTxt = static_cast<CEGUI::StaticText*>(wmgr.getWindow((CEGUI::utf8*)"OctavesTxt"));
		mZoomTxt = static_cast<CEGUI::StaticText*>(wmgr.getWindow((CEGUI::utf8*)"ZoomTxt"));
		mOffsetTxt = static_cast<CEGUI::StaticText*>(wmgr.getWindow((CEGUI::utf8*)"OffsetTxt"));
		mGainTxt = static_cast<CEGUI::StaticText*>(wmgr.getWindow((CEGUI::utf8*)"GainTxt"));

		mBasis1 = static_cast<CEGUI::RadioButton*>(wmgr.getWindow((CEGUI::utf8*)"Basis1"));
		mBasis2 = static_cast<CEGUI::RadioButton*>(wmgr.getWindow((CEGUI::utf8*)"Basis2"));
		mBasis3 = static_cast<CEGUI::RadioButton*>(wmgr.getWindow((CEGUI::utf8*)"Basis3"));
		mBasis4 = static_cast<CEGUI::RadioButton*>(wmgr.getWindow((CEGUI::utf8*)"Basis4"));
		mBasis5 = static_cast<CEGUI::RadioButton*>(wmgr.getWindow((CEGUI::utf8*)"Basis5"));
		mBasis6 = static_cast<CEGUI::RadioButton*>(wmgr.getWindow((CEGUI::utf8*)"Basis6"));
      
        CEGUI::Point radioPos = mBasis6->getPosition();
        const float height = mBasis6->getHeight();

		mBasis7 = static_cast<CEGUI::RadioButton*>(wmgr.getWindow((CEGUI::utf8*)"Basis7"));
        radioPos.d_y += height;
        mBasis7->setPosition(radioPos);

		mBasis8 = static_cast<CEGUI::RadioButton*>(wmgr.getWindow((CEGUI::utf8*)"Basis8"));
        radioPos.d_y += height;
        mBasis8->setPosition(radioPos);

		mBasis9 = static_cast<CEGUI::RadioButton*>(wmgr.getWindow((CEGUI::utf8*)"Basis9"));
        radioPos.d_y += height;
        mBasis9->setPosition(radioPos);

		mBasis10 = static_cast<CEGUI::RadioButton*>(wmgr.getWindow((CEGUI::utf8*)"Basis10"));
        radioPos.d_y += height;
        mBasis10->setPosition(radioPos);

        mBasis11 = static_cast<CEGUI::RadioButton*>(wmgr.getWindow((CEGUI::utf8*)"Basis11"));
        radioPos.d_y += height;
        mBasis11->setPosition(radioPos);

		mSeedTxt = static_cast<CEGUI::Editbox*>(wmgr.getWindow((CEGUI::utf8*)"SeedTxt"));
		mUpBtn = static_cast<CEGUI::PushButton*>(wmgr.getWindow((CEGUI::utf8*)"UpBtn"));
		mDownBtn = static_cast<CEGUI::PushButton*>(wmgr.getWindow((CEGUI::utf8*)"DownBtn"));
		mLeftBtn = static_cast<CEGUI::PushButton*>(wmgr.getWindow((CEGUI::utf8*)"LeftBtn"));
		mRightBtn = static_cast<CEGUI::PushButton*>(wmgr.getWindow((CEGUI::utf8*)"RightBtn"));
		mFilenameTxt = static_cast<CEGUI::Editbox*>(wmgr.getWindow((CEGUI::utf8*)"FilenameTxt"));
		mSaveinfoTxt = static_cast<CEGUI::StaticText*>(wmgr.getWindow((CEGUI::utf8*)"SaveinfoTxt"));

		// Scrollbar properties
		mScaleScr->setStepSize(0.01);
		mOctavesScr->setStepSize(0.05);
		mZoomScr->setStepSize(1.0/16.0);
		mOffsetScr->setStepSize(0.005);
		mGainScr->setStepSize(0.01);

		// Default values
		mScaleScr->setScrollPosition(0.1);
		mOctavesScr->setScrollPosition(0.4);
		mZoomScr->setScrollPosition(10.0/16.0);	// 1024
		mOffsetScr->setScrollPosition(0.57/2.0);
		mGainScr->setScrollPosition(3.5/10.0);

		// Update Labels
		ScrollValues();

	}

    void setupEventHandlers(void)
    {
		CEGUI::WindowManager& wmgr = CEGUI::WindowManager::getSingleton();
        wmgr.getWindow((CEGUI::utf8*)"ExitBtn")
			->subscribeEvent(
				CEGUI::PushButton::EventClicked, 
				CEGUI::Event::Subscriber(&GuiApplication::handleQuit, this));
        wmgr.getWindow((CEGUI::utf8*)"ScaleScr")
			->subscribeEvent(
				CEGUI::Scrollbar::EventScrollPositionChanged, 
				CEGUI::Event::Subscriber(&GuiApplication::handleChangeScroll, this));
        wmgr.getWindow((CEGUI::utf8*)"OctavesScr")
			->subscribeEvent(
				CEGUI::Scrollbar::EventScrollPositionChanged, 
				CEGUI::Event::Subscriber(&GuiApplication::handleChangeScroll, this));
        wmgr.getWindow((CEGUI::utf8*)"ZoomScr")
			->subscribeEvent(
				CEGUI::Scrollbar::EventScrollPositionChanged, 
				CEGUI::Event::Subscriber(&GuiApplication::handleChangeScroll, this));
        wmgr.getWindow((CEGUI::utf8*)"OffsetScr")
			->subscribeEvent(
				CEGUI::Scrollbar::EventScrollPositionChanged, 
				CEGUI::Event::Subscriber(&GuiApplication::handleChangeScroll, this));
        wmgr.getWindow((CEGUI::utf8*)"GainScr")
			->subscribeEvent(
				CEGUI::Scrollbar::EventScrollPositionChanged, 
				CEGUI::Event::Subscriber(&GuiApplication::handleChangeScroll, this));
        wmgr.getWindow((CEGUI::utf8*)"UpdateBtn")
			->subscribeEvent(
				CEGUI::PushButton::EventClicked, 
				CEGUI::Event::Subscriber(&GuiApplication::handleUpdate, this));
        wmgr.getWindow((CEGUI::utf8*)"UpBtn")
			->subscribeEvent(
				CEGUI::PushButton::EventClicked, 
				CEGUI::Event::Subscriber(&GuiApplication::handlePan, this));
        wmgr.getWindow((CEGUI::utf8*)"DownBtn")
			->subscribeEvent(
				CEGUI::PushButton::EventClicked, 
				CEGUI::Event::Subscriber(&GuiApplication::handlePan, this));
        wmgr.getWindow((CEGUI::utf8*)"LeftBtn")
			->subscribeEvent(
				CEGUI::PushButton::EventClicked, 
				CEGUI::Event::Subscriber(&GuiApplication::handlePan, this));
        wmgr.getWindow((CEGUI::utf8*)"RightBtn")
			->subscribeEvent(
				CEGUI::PushButton::EventClicked, 
				CEGUI::Event::Subscriber(&GuiApplication::handlePan, this));
        wmgr.getWindow((CEGUI::utf8*)"SaveBtn")
			->subscribeEvent(
				CEGUI::PushButton::EventClicked, 
				CEGUI::Event::Subscriber(&GuiApplication::handleSave, this));
        wmgr.getWindow((CEGUI::utf8*)"FilenameTxt")
			->subscribeEvent(
				CEGUI::Editbox::EventActivated, 
				CEGUI::Event::Subscriber(&GuiApplication::handleEdit, this));
        wmgr.getWindow((CEGUI::utf8*)"SeedTxt")
			->subscribeEvent(
				CEGUI::Editbox::EventActivated, 
				CEGUI::Event::Subscriber(&GuiApplication::handleEdit, this));
    }

	bool handleEdit(const CEGUI::EventArgs& e)
    {
		// We need to switch back to buffered key input here
		GuiFrameListener* pek = (GuiFrameListener*)mFrameListener;
		pek->setKeyInputType(true);
		return true;
	}

	bool handleSave(const CEGUI::EventArgs& e)
    {
		const String fname = mFilenameTxt->getText().c_str();
		MeshSerializer *msave = new MeshSerializer();
		MeshPtr pterrmesh = mSceneMgr->getEntity("plane02")->getMesh();
		msave->exportMesh(pterrmesh.get(), fname);
		pterrmesh.setNull();
		delete msave;

        String ImageName;
        std::vector<String> stvec = StringUtil::split(fname, ".");
        if (stvec.size() > 1)
        {
            if (stvec.size() > 3)
                ImageName = stvec[0] + stvec[1] + ".png";
            else
                ImageName = "." + stvec[0] + ".png";
        }
        else
            ImageName = ImageName + ".png";

        testft->SaveMap("plane02", ImageName);

        
        ImageName = "." + stvec[0] + ".raw";
        testft->SaveRawMap("plane02", ImageName);

		return true;
	}

	bool handlePan(const CEGUI::EventArgs& e)
    {
        const CEGUI::WindowEventArgs& we = static_cast<const CEGUI::WindowEventArgs&>(e);
		if(we.window->getName() == "UpBtn")
			testft->AddZoff(-128.0);
		if(we.window->getName() == "DownBtn")
			testft->AddZoff(128.0);
		if(we.window->getName() == "LeftBtn")
			testft->AddXoff(-128.0);
		if(we.window->getName() == "RightBtn")
			testft->AddXoff(128.0);

		testft->Update("plane02", TERRAIN_SIZE);
		return true;
	}

	bool handleUpdate(const CEGUI::EventArgs& e)
    {
		float foffset, fgain;
		int iscale, ioctaves;
		int iseed, izoom, ival = 0;

		// Read values
        iscale = (int) (mScaleScr->getScrollPosition() * 1999 + 1);
        izoom = pow(2.0f, (long) (mZoomScr->getScrollPosition() * 15 + 1));
		ioctaves = (int) (mOctavesScr->getScrollPosition() * 19 + 1);
		foffset = mOffsetScr->getScrollPosition() * 2.0;
		fgain = mGainScr->getScrollPosition() * 10.0;
		iseed = atoi(mSeedTxt->getText().c_str());

		if(mBasis1->isSelected())	ival = 0;
		if(mBasis2->isSelected())	ival = 1;
		if(mBasis3->isSelected())	ival = 2;
		if(mBasis4->isSelected())	ival = 3;
		if(mBasis5->isSelected())	ival = 4;
		if(mBasis6->isSelected())	ival = 5;

		if(mBasis7->isSelected())	ival = 6;
		if(mBasis8->isSelected())	ival = 7;
		if(mBasis9->isSelected())	ival = 8;
		if(mBasis10->isSelected())	ival = 9;
        if(mBasis11->isSelected())	ival = 10;

		// Update terrain
		testft->SetYscale(iscale);
		testft->SetZoom(izoom);
		testft->SetOctaves(ioctaves);
		testft->SetOffset(foffset);
		testft->SetGain(fgain);
		testft->SetBasisFunc(ival);
		testft->SetSeed(iseed);

        if (ival == 10)
            testft->UpdateWaterErode("plane02", TERRAIN_SIZE);
        else if (ival > 5)
		    testft->UpdateFault("plane02", TERRAIN_SIZE);
        else
		    testft->Update("plane02", TERRAIN_SIZE);


		return true;
	}

    bool handleChangeScroll(const CEGUI::EventArgs& e)
	{
		ScrollValues();
		return true;
	}

    void ScrollValues()
    {
		String stext;
		float fscale, foctaves, fzoom, foffset, fgain;
		int iscale, ioctaves, izoom;
		CEGUI::WindowManager& wmgr = CEGUI::WindowManager::getSingleton();

		// Decode Scale parameter
        fscale = mScaleScr->getScrollPosition() * 1999 + 1;
		iscale = (int) fscale;
		stext = "Scale: ";
		stext += MY_NUM(iscale);
		mScaleTxt->setText(stext);
		// Decode Zoom parameter
        fzoom = mZoomScr->getScrollPosition() * 15 + 1;
		izoom = (int) fzoom;
		izoom = pow(2.0f, (long)izoom);
		stext = "Zoom: ";
		stext += MY_NUM(izoom);
		mZoomTxt->setText(stext);
		// Decode Octaves parameter
        foctaves = mOctavesScr->getScrollPosition() * 19 + 1;
		ioctaves = (int) foctaves;
		stext = "Octaves: ";
		stext += MY_NUM(ioctaves);
		mOctavesTxt->setText(stext);
		// Decode Offset parameter
        foffset = mOffsetScr->getScrollPosition() * 2.0;
		stext = "Offset: ";
		stext += MY_NUM(foffset);
		mOffsetTxt->setText(stext);
		// Decode Gain parameter
        fgain = mGainScr->getScrollPosition() * 10.0;
		stext = "Gain: ";
		stext += MY_NUM(fgain);
		mGainTxt->setText(stext);
	}

	bool handleQuit(const CEGUI::EventArgs& e)
    {
        static_cast<GuiFrameListener*>(mFrameListener)->requestShutdown();
        return true;
    }

	bool handleMouseEnters(const CEGUI::EventArgs& e)
	{
		CEGUI::WindowEventArgs& we = ((CEGUI::WindowEventArgs&)e);		
		return true;
	}
	bool handleMouseLeaves(const CEGUI::EventArgs& e)
	{
		return true;
	}

};

#if OGRE_PLATFORM == OGRE_PLATFORM_WIN32
#define WIN32_LEAN_AND_MEAN
#include "windows.h"

INT WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR strCmdLine, INT )
#else
int main(int argc, char *argv[])
#endif
{

    // Create application object
    GuiApplication app;

    try {
        app.go();
    } catch( Ogre::Exception& e ) {
#if OGRE_PLATFORM == OGRE_PLATFORM_WIN32
        MessageBox( NULL, e.getFullDescription().c_str(), "An exception has occured!", MB_OK | MB_ICONERROR | MB_TASKMODAL);
#else
        std::cerr << "An exception has occured: " <<
            e.getFullDescription().c_str() << std::endl;
#endif
    }


    return 0;
}

