
#include "Ogre.h"
#include "OgreMovableObject.h"
#include "OgreSimpleRenderable.h"

#include "TerrainMesh.h"


namespace Ogre
{
    //-----------------------------------------------------------------------
    TerrainMesh::TerrainMesh(const String& MeshName, const String& MaterialName): 
        meshName(MeshName),
        mMaterialName(MaterialName),
        mVertexBuffer (0),
        mNormalBuffer (0),
        mTexcoordsBuffer (0)
    {    
        mMesh.setNull();  
    }
    //-----------------------------------------------------------------------
    TerrainMesh::~TerrainMesh()
    {
        mMesh.setNull();
    }
    //-----------------------------------------------------------------------
    void TerrainMesh::setMaterialName(const String &m)
    {
        mMaterialName = m;
    }
    //-----------------------------------------------------------------------
    bool TerrainMesh::init(const uint numvertX, 
                           const uint numvertZ, 
                           const Real TerrainSizeX, 
                           const Real TerrainSizeZ)
    {
        mNumvertX = numvertX + 1;
        mNumvertZ = numvertZ + 1;  

        mScaleX = TerrainSizeX / numvertX;
        mScaleZ = TerrainSizeZ / numvertZ;


        initBuffers ();

        mInit = true;
        return true;
    }

    bool TerrainMesh::update(const float * const heightData)
    {
		// Update position array
		float miny = 0.0f, maxy = 0.0f;
        const float xscale = mScaleX;
        const float zscale = mScaleZ;

		HardwareVertexBufferSharedPtr vbuf = mVertexData->vertexBufferBinding->getBuffer(TERRAINMESH_VERTEX_BINDING);

		float *vDest = static_cast<float*>(vbuf->lock(HardwareBuffer::HBL_DISCARD));
        float z_offset = 0.0f;
        uint z_offset_data = 0;
        for (uint z = 0; z < mNumvertZ; ++z )
        {
            float x_offset = 0.0f;
            for (uint x = 0; x < mNumvertX; ++x )
            {
                // update positions
			    const float y = heightData[z_offset_data + x];

			    if(y > maxy) 
                    maxy = y;
			    if(y < miny) 
                    miny = y;

			    *(vDest++) = x_offset;
			    *(vDest++) = y;
			    *(vDest++) = z_offset;

                 x_offset += xscale;

            } // for (uint x = 0; x < side; ++x ) 
            z_offset += zscale;
            z_offset_data += mNumvertX;
		}
		vbuf->unlock();

        calculateNormals(heightData);

		AxisAlignedBox mmbox = mMesh->getBounds();
		mmbox.merge(Vector3(0.0f, maxy, 0.0f));
		mMesh->_setBounds(mmbox);
        return true;
	}
    //-----------------------------------------------------------------------
    void TerrainMesh::calculateNormals( const float * const heightmap)
    {
        
       const uint numvertZ = mNumvertZ;
       const uint numvertX = mNumvertX;

	   HardwareVertexBufferSharedPtr nbuf = mVertexData->vertexBufferBinding->getBuffer(TERRAINMESH_NORMAL_BINDING);
	   float *nDest = static_cast<float*>(nbuf->lock(HardwareBuffer::HBL_DISCARD));

        memset (nDest, 0, numvertZ*numvertX*sizeof(float)*3);

       const uint width = numvertX;
       uint curr_row = width;
       nDest += width*3;
       Vector3 Normal;
       for(uint z = 1; z < numvertZ - 1; ++z) 
        {
            nDest += 3;
		    for (uint x = 1; x < width - 1; ++x) 
            {
                // Fast SOBEL filter and mathematically correct
                
                // x
                Normal.x = heightmap[x-1 + curr_row-width] + heightmap[x-1 + curr_row]
                         + heightmap[x-1 + curr_row]       + heightmap[x-1 + curr_row+width]
                         - heightmap[x+1 + curr_row-width] - heightmap[x+1 + curr_row] 
                         - heightmap[x+1 + curr_row]       - heightmap[x+1 + curr_row+width];

                // y
                Normal.y = 8.0f;

                // z
                Normal.z = heightmap[x-1 + curr_row-width] + heightmap[x +   curr_row-width]
                         + heightmap[x +   curr_row-width] + heightmap[x+1 + curr_row-width]
                         - heightmap[x-1 + curr_row+width] - heightmap[x +   curr_row+width]
                         - heightmap[x +   curr_row+width] - heightmap[x+1 + curr_row+width];

                            
                Normal.normalise ();

                *(nDest++) = Normal.x;
                *(nDest++) = Normal.y;
                *(nDest++) = Normal.z;
            } // for (uint x = 1; x < width - 1; ++x)
            nDest += 3;
            curr_row += width;
        } // for(uint z = 1; z < numvertZ - 1; ++z)
        nbuf->unlock();
    }
    //-----------------------------------------------------------------------
    void TerrainMesh::initBuffers()
    {
        // create mesh and sub-meshes
        mMesh = MeshManager::getSingleton().createManual(meshName, 
            ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME) ;

		//Init index buffer
		mVertexCount = (mNumvertX + 1) * (mNumvertZ + 1);  
        const uint indexCount = 3 * 2 * (mNumvertX-1)*(mNumvertZ-1);
        mNumFaces = indexCount / 3;
        HardwareIndexBufferSharedPtr indexHardBuffer;
        if (mVertexCount < 65535)
        {
            indexHardBuffer = 
                            HardwareBufferManager::getSingleton().createIndexBuffer(
                            HardwareIndexBuffer::IT_16BIT, 
                            indexCount, 
                            HardwareBuffer::HBU_STATIC, 
                            false); // use shadow buffer

           ushort *Indexes = new ushort [indexCount];
            
            // multi Index Level
            ushort *pIdx = Indexes;
            uint height_offset = 0;
            for (uint j = 0; j < mNumvertZ - 1; ++j )
            {
                for (uint i = 0; i < mNumvertX - 1; ++i )
                {
                    //triangles
                    *pIdx++ = i + height_offset;       
                    *pIdx++ = i + height_offset + mNumvertX; 
                    *pIdx++ = i + 1 + height_offset; 

                    *pIdx++ = i + height_offset + mNumvertX;
                    *pIdx++ = i + 1 + height_offset + mNumvertX; 
                    *pIdx++ = i + 1 + height_offset;      
                 }
                height_offset += mNumvertX;
            }
            indexHardBuffer->writeData(0, 
                sizeof(ushort) * indexCount, // size 
                Indexes, // source
                true); // discard?  
            delete[] Indexes;

        } // if (mIndexCount < 65535)
        else
        {
            indexHardBuffer = 
                            HardwareBufferManager::getSingleton().createIndexBuffer(
                            HardwareIndexBuffer::IT_32BIT, 
                            indexCount, 
                            HardwareBuffer::HBU_STATIC, 
                            false); // use shadow buffer
            uint *Indexes = new uint [indexCount];
              // multi Index Level
            uint *pIdx = Indexes;
            uint height_offset = 0;
            for (uint j = 0; j < mNumvertZ - 1; ++j )
            {
                for (uint i = 0; i < mNumvertX - 1; ++i )
                {
                    //triangles
                    *pIdx++ = i + height_offset;       
                    *pIdx++ = i + height_offset + mNumvertX; 
                    *pIdx++ = i + 1 + height_offset; 

                    *pIdx++ = i + height_offset + mNumvertX;
                    *pIdx++ = i + 1 + height_offset + mNumvertX; 
                    *pIdx++ = i + 1 + height_offset;      
                 }
                height_offset += mNumvertX;
            }
            indexHardBuffer->writeData(0, 
                sizeof(uint) * indexCount, // size 
                Indexes, // source
                true); // discard?  
            delete[] Indexes;
        }

        IndexData*  mIndexData = new IndexData;
        mIndexData->indexStart = 0;
        mIndexData->indexCount = indexCount;
        mIndexData->indexBuffer = indexHardBuffer;


        //Init vertex buffer
        mVec3Buffsize = sizeof(Vector3) * mVertexCount;

        mVertexData= new VertexData;
        mVertexData->vertexStart = 0;
        mVertexData->vertexCount = mVertexCount;
        
        VertexDeclaration* decl = mVertexData->vertexDeclaration;
        VertexBufferBinding* bind = mVertexData->vertexBufferBinding;

        decl->addElement(TERRAINMESH_VERTEX_BINDING, 0, VET_FLOAT3, VES_POSITION);
        decl->addElement(TERRAINMESH_NORMAL_BINDING, 0, VET_FLOAT3, VES_NORMAL);
        decl->addElement(TERRAINMESH_TEXCOORDBINDING, 0, VET_FLOAT2, VES_TEXTURE_COORDINATES);


        HardwareVertexBufferSharedPtr posVertexHardBuffer;
        HardwareVertexBufferSharedPtr texcoordHardBuffer;
        HardwareVertexBufferSharedPtr normVertexHardBuffer;

        // Create shared vertex buffer
        posVertexHardBuffer =
            HardwareBufferManager::getSingleton().createVertexBuffer(
            sizeof(Vector3),//decl->getVertexSize(VERTEX_BINDING),
            mVertexCount, 
            //HardwareBuffer::HBU_DYNAMIC_WRITE_ONLY);
            HardwareBuffer::HBU_STATIC_WRITE_ONLY,
            false); // use shadow buffer
        bind->setBinding(TERRAINMESH_VERTEX_BINDING, posVertexHardBuffer);

        // Prepare buffer for normals - write only
        normVertexHardBuffer = 
            HardwareBufferManager::getSingleton().createVertexBuffer( 
            sizeof(Vector3),//decl->getVertexSize(NORMAL_BINDING), 
            mVertexCount, 
            HardwareBuffer::HBU_STATIC_WRITE_ONLY,
            false); // use shadow buffer
        bind->setBinding(TERRAINMESH_NORMAL_BINDING, normVertexHardBuffer);

        texcoordHardBuffer =
            HardwareBufferManager::getSingleton().createVertexBuffer(
            sizeof(Vector2),//decl->getVertexSize(TEXCOORDBINDING),
            mVertexCount, 
            HardwareBuffer::HBU_STATIC_WRITE_ONLY,
            false); // use shadow buffer
        bind->setBinding(TERRAINMESH_TEXCOORDBINDING, texcoordHardBuffer);


        //Init vertex buffer:
        mVertexBuffer = new Vector3[mVertexCount];
        mNormalBuffer = new Vector3[mVertexCount];
        mTexcoordsBuffer = new Vector2[mVertexCount];

        uint ArrayPos = 0;
        Real z_distance = 0.0f;
        for (uint j = 0; j < mNumvertZ; j++)
        {
            Real x_distance = 0.0f;
            const Real z_coord = static_cast<Real> (j)  / (mNumvertZ - 1);
            for (uint i = 0; i < mNumvertX; i++)
            {
                mVertexBuffer[ArrayPos] = Vector3 (x_distance,
                                                   0.0f,
                                                   z_distance);
                mNormalBuffer[ArrayPos] = Vector3::NEGATIVE_UNIT_Y;
                mTexcoordsBuffer[ArrayPos] = Vector2 (1 - (static_cast<Real> (i)  / (mNumvertX - 1)),
                                                     1 - z_coord);

                x_distance += mScaleX;
                ArrayPos++;
            }
            z_distance += mScaleZ;
        }

        posVertexHardBuffer->writeData(0, 
            mVec3Buffsize, // size 
            mVertexBuffer, // source
            true); // discard?

        normVertexHardBuffer->writeData(0, 
            mVec3Buffsize, // size 
            mNormalBuffer, // source
            true); // discard?

        texcoordHardBuffer->writeData(0, 
            sizeof(Vector2) * mVertexCount, // size 
            mTexcoordsBuffer, // source
            true); // discard?

        delete[] mVertexBuffer;
        delete[] mNormalBuffer;
        delete[] mTexcoordsBuffer;


        mTerrainMesh = mMesh->createSubMesh();
        mTerrainMesh->useSharedVertices = false;
        mTerrainMesh->vertexData = mVertexData;
        mTerrainMesh->indexData = mIndexData;

        
        mTerrainMesh->setMaterialName (mMaterialName);

        mTerrainMesh->operationType = RenderOperation::OT_TRIANGLE_LIST;

        AxisAlignedBox meshBounds(0.0f, 
                                  0.0f, 
                                  0.0f,
                                  mNumvertX*mScaleX, 
                                  1.0f, 
                                  mNumvertZ*mScaleZ);
        mMesh->_setBounds(meshBounds);

        mMesh->load();
        mMesh->touch();

    }   
}
