/***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU Lesser General Public License as        *
*   published by the Free Software Foundation; either version 2 of the    *
*   License, or (at your option) any later version.                       *
*                                                                         *
***************************************************************************/
/**
@file 
WaterErode.cpp
@brief
Code that process a water erosion iterative model on a heightmap
@author
sputnick
@date
25/mar/2005

*/

#include "WaterErode.h"


namespace Ogre
{
    //-----------------------------------------------------------------------
    
    WaterErode::WaterErode()
    {
    }
    
    //-----------------------------------------------------------------------
    
    WaterErode::~WaterErode()
    { 
    }
    
    //-----------------------------------------------------------------------
  
    void WaterErode::getNeighborsDelta(Real* const afMap, const uint size, uint u32Offset, double *adNeighborsDelta) const
    {
        adNeighborsDelta[0] = afMap[u32Offset - size - 1] - afMap[u32Offset];
        adNeighborsDelta[1] = afMap[u32Offset - size] - afMap[u32Offset];
        adNeighborsDelta[2] = afMap[u32Offset - size + 1] - afMap[u32Offset];
        adNeighborsDelta[3] = afMap[u32Offset - 1] - afMap[u32Offset];
        adNeighborsDelta[4] = afMap[u32Offset + 1] - afMap[u32Offset];
        adNeighborsDelta[5] = afMap[u32Offset + size - 1] - afMap[u32Offset];
        adNeighborsDelta[6] = afMap[u32Offset + size] - afMap[u32Offset];
        adNeighborsDelta[7] = afMap[u32Offset + size + 1] - afMap[u32Offset];
    }

    //---------------------------------------------------------------------------

    void WaterErode::Erode(Real* const afMap, const uint size,const uint iNbIteration) const
    {
        Real *afWaterMap, *afSedimentMap, *afAltMap, *afNewWaterMap, *afNewSedimentMap, *afNewAltMap;
        Real *pTempMap;
        uint i, j, x, y, u32Offset;
   

        afWaterMap = new Real[size*size];
        afSedimentMap = new Real[size*size];
        afAltMap = new Real[size*size];
        afNewWaterMap = new Real[size*size];
        afNewSedimentMap = new Real[size*size];
        afNewAltMap = new Real[size*size];

        double dKd = 0.05;  // 0.1
        double dKc = 1.0;  // 2.0
        double dKs = 0.025; // 0.05
        double dSedCap;

      //  double dMaxWater = afMap->getMaxValue(true);
      //  double dMinVal = afMap->getMinValue(true);

        double dDeltaW;
        double adNeighborsWeights[8];
        double dNeighborsWeightsSum;
        int    aiOffsetArray[8];
        // int    iOffset;
        int    iCount;

        aiOffsetArray[0] = -(int)size - 1;
        aiOffsetArray[1] = -(int)size;
        aiOffsetArray[2] = -(int)size + 1;
        aiOffsetArray[3] = -1;
        aiOffsetArray[4] = 1;
        aiOffsetArray[5] = size - 1;
        aiOffsetArray[6] = size;
        aiOffsetArray[7] = size + 1;

        u32Offset = 0;
        for (y=0; y<size; y++)
          for (x=0; x<size; x++)
          {
            afWaterMap[u32Offset] = 0.5;
            afSedimentMap[u32Offset] = 0.0;
            afAltMap[u32Offset] = afMap[u32Offset];
            u32Offset++;
          }

        for (i=0; i<iNbIteration; i++)
        {
          u32Offset = 0;
          for (y=0; y<size; y++)
            for (x=0; x<size; x++)
            {
              afNewSedimentMap[u32Offset] = 0.0;
              afNewWaterMap[u32Offset] = 0.0;
              afNewAltMap[u32Offset] = 0.0;
              u32Offset++;
            }

          u32Offset = size*(size-1);
          for (x=0; x<size; x++)
          {
            afNewAltMap[x] = afAltMap[x];
            afNewAltMap[u32Offset+x] = afAltMap[u32Offset+x];
          }

          u32Offset = 0;
          for (y=0; y<size; y++)
          {
            afNewAltMap[u32Offset] = afAltMap[u32Offset];
            afNewAltMap[u32Offset+size-1] =afAltMap[u32Offset+size-1];
            u32Offset += size;
          }



          for (y=1; y<size-1; y++)
          {
            u32Offset = y*size + 1;
            for (x=1; x<size-1; x++)
            {
   //           u32Offset = y*size + x;
              getNeighborsDelta(afAltMap, size, u32Offset, adNeighborsWeights);

              dNeighborsWeightsSum = 0.0;
              iCount = 0;
              for (j=0; j<8; j++)
              {
                if (adNeighborsWeights[j] < 0.0)
                {
                  dNeighborsWeightsSum -= adNeighborsWeights[j];
                  iCount++;
                }
              }

              if (dNeighborsWeightsSum <= 0.0)
              {
                afNewAltMap[u32Offset] = afAltMap[u32Offset];
                afNewWaterMap[u32Offset] = afWaterMap[u32Offset];
                afNewSedimentMap[u32Offset] = afSedimentMap[u32Offset];
              }
              else
              {
                for (j=0; j<8; j++)
                {
                  if (adNeighborsWeights[j] < 0.0)
                  {
                    adNeighborsWeights[j] /= dNeighborsWeightsSum;
                    dDeltaW = afWaterMap[u32Offset] -  afWaterMap[(int)u32Offset+aiOffsetArray[j]];
                    dDeltaW += afAltMap[u32Offset] -  afAltMap[(int)u32Offset+aiOffsetArray[j]];
            //        dDeltaW *= -adNeighborsWeights[j];
                    if (afWaterMap[u32Offset] < dDeltaW)
                      dDeltaW = afWaterMap[u32Offset];

                    if (dDeltaW <= 0.0) // depose sediment
                    {
                      afNewAltMap[u32Offset] += -adNeighborsWeights[j]*(afAltMap[u32Offset] + dKd*afSedimentMap[u32Offset]);
                      afNewSedimentMap[u32Offset] += -adNeighborsWeights[j]*(afSedimentMap[u32Offset] * (1 - dKd));
                    }
                    else  // flow water
                    {
                      afNewWaterMap[u32Offset] += -adNeighborsWeights[j]*(afWaterMap[u32Offset] - dDeltaW);
                      afNewWaterMap[(int)u32Offset+aiOffsetArray[j]] += -adNeighborsWeights[j]*(afWaterMap[(int)u32Offset+aiOffsetArray[j]] + dDeltaW);
                      dSedCap = dKc*dDeltaW;
                      if (afSedimentMap[u32Offset] >= dSedCap)
                      {
                        afNewSedimentMap[(int)u32Offset+aiOffsetArray[j]] += -adNeighborsWeights[j]*(afSedimentMap[(int)u32Offset+aiOffsetArray[j]] + dSedCap);
                        afNewAltMap[u32Offset] += -adNeighborsWeights[j]*(afAltMap[u32Offset] + dKd*(afSedimentMap[u32Offset] - dSedCap));
                        afNewSedimentMap[u32Offset] += -adNeighborsWeights[j]*(1.0 - dKd)*(afSedimentMap[u32Offset] - dSedCap);
                      }
                      else
                      {
                        afNewSedimentMap[(int)u32Offset+aiOffsetArray[j]] += -adNeighborsWeights[j]*(afSedimentMap[(int)u32Offset+aiOffsetArray[j]] + afSedimentMap[u32Offset] + dKs*(dSedCap - afSedimentMap[u32Offset]));
                        afNewAltMap[u32Offset] += -adNeighborsWeights[j]*(afAltMap[u32Offset] - dKs*(dSedCap - afSedimentMap[u32Offset]));
                    //    afNewSedimentMap[u32Offset] += 0.0;
                      }

                    }

                  }

                }
              }

               u32Offset++;
            }
          }


          pTempMap = afWaterMap;
          afWaterMap = afNewWaterMap;
          afNewWaterMap = pTempMap;

          pTempMap = afSedimentMap;
          afSedimentMap = afNewSedimentMap;
          afNewSedimentMap = pTempMap;

          pTempMap = afAltMap;
          afAltMap = afNewAltMap;
          afNewAltMap = pTempMap;

      /////////////

          u32Offset = 0;
          for (y=0; y<size; y++)
            for (x=0; x<size; x++)
            {
              afMap[u32Offset] = afAltMap[u32Offset];
              u32Offset++;
            }
//          refreshDisplay();


      /////////////
        }

      //  u32Offset = 0;
        for (y=1; y<size-1; y++)
          for (x=1; x<size-1; x++)
          {

#if 1
            afMap[x+y*size] = afAltMap[x-1+(y-1)*size]
                                          + 2*afAltMap[x+(y-1)*size]
                                          + afAltMap[x+1+(y-1)*size]
                                          + 2*afAltMap[x-1+y*size]
                                          + 4*afAltMap[x+y*size]
                                          + 2*afAltMap[x+1+y*size]
                                          + afAltMap[x-1+(y+1)*size]
                                          + 2*afAltMap[x+(y+1)*size]
                                          + afAltMap[x+1+(y+1)*size];

            afMap[x+y*size] /= 16;
#else
            afMap[x+y*size] = afAltMap[x+y*size];
#endif

          }

        u32Offset = size*(size-1);
        for (x=0; x<size; x++)
        {
          afMap[x] = afAltMap[x];
          afMap[u32Offset+x] = afAltMap[u32Offset+x];
        }
        u32Offset = 0;
        for (y=0; y<size; y++)
        {
          afMap[u32Offset] = afAltMap[u32Offset];
          afMap[u32Offset+size-1] = afAltMap[u32Offset+size-1];
          u32Offset += size;
        }



        delete[] afWaterMap;
        delete[] afSedimentMap;
        delete[] afAltMap;
        delete[] afNewWaterMap;
        delete[] afNewSedimentMap;
        delete[] afNewAltMap;
    }
    
    //-----------------------------------------------------------------------
 
    void WaterErode::normalize (Real * const field, const uint size,
                                const float scale) const
    {
        /*
	    Find the maximum and minimum values in the height field
	    */ 
        const uint fieldSize = size*size;
	    Real maxVal = field[0];
	    Real minVal = field[0];
        uint i;
	    for (i = 1; i < fieldSize; i++)
	    {
		    if (field[i] > maxVal) 
		    {
			    maxVal = field[i];
		    }
		    else if (field[i] < minVal) 
		    {
			    minVal = field[i];
		    }
	    }

	    /*
	    Find the altitude range (dh)
	    */
        const Real minFinal = minVal;
	    if (maxVal <= minFinal) 
            return;
	    const Real inv_dh = scale / (maxVal - minFinal);

	    /*
	    Scale all the values so they fill exactly the range 0-1
	    */
	    for (i=0; i < fieldSize; i++)
	    {
		    field[i] = (field[i] - minFinal) * inv_dh;
            assert (field[i] <= 1.0f && field[i] >= 0.0f);
	    }
    } 
    
    //-----------------------------------------------------------------------
 
    void WaterErode::equalize (Real * const field, const uint size, 
                            const Real hmin, const Real hmax, 
                            const uint maxColor) const
    {
        //maxColor  is 65535 or 255;

        const double divider = 1.0f / (hmax-hmin) * maxColor;
        const uint data_size = size*size;
        uchar *uData = new uchar[data_size];
        uint i;
        for (i = 0; i < data_size; i++ )            
        { 
            uData[i] = uchar  ((field[i]-hmin) * divider);
        }


        uint *freq = new uint[maxColor];
        memset (freq, 0, maxColor*sizeof(uint));
        for (i = 0; i < size; i++ )            
        { 
                freq[uData[i]] += 1;
        }

        uint *Cumulfreq = freq;
        for (i = 1; i < maxColor; i++)
        {           
            Cumulfreq[i] = Cumulfreq[i - 1] + freq[i];
        }

        const double factor = static_cast <double> (hmax - hmin) / data_size;
        uint cur_pos = 0;
        for (i = 0; i < data_size; i++)
        {
            field[i] = Cumulfreq[uData[i]] * factor;           
        }
        delete[] freq;
    }
     
    //-----------------------------------------------------------------------
    
}//namespace Ogre
